unit uCiaPipes;

interface

uses
  Windows, Classes, SysUtils, Messages;

const
  WM_DATAAVAILABLE = WM_USER + 1;
  WM_SESSION_CONNECTED = WM_USER + 2;
  WM_SESSION_DISCONNECTED = WM_USER + 3;
  WM_ERROR = WM_USER + 4;

type
  TPipeError = procedure(Sender: TObject; Error: integer) of object;
  TPipeDataAvailable = procedure(Sender: TObject; Data: PChar; Len: integer) of object;
  TPipeDataSent = procedure(Sender: TObject; BytesTransferred: integer) of object;
  EPipe = class(Exception);
  TPipe = class(TThread)
  private
    FBuffer: PChar;
    FBufSize: cardinal;
    CloseEvent: THandle;
    Handle: Hwnd;
    FPipeHandle: THandle;
    FOnDataAvailable: TPipeDataAvailable;
    FPipeName: string;
    FOnSessionDisconnected: TNotifyEvent;
    FOnSessionConnected: TNotifyEvent;
    FOnError: TPipeError;
    FOnDataSent: TPipeDataSent;
    procedure TriggerError(Error: integer);
    procedure TriggerSessionConnected;
    procedure TriggerSessionDisconnected;
    procedure TriggerDataAvailable(Data: PChar; Len: integer);
    procedure TriggerDataSent(BytesTransferred: integer);
    procedure WndProc(var MsgRec: TMessage);
  protected
    procedure Execute; override;
  public
    constructor create(const PipeName: string); reintroduce; virtual;
    destructor  Destroy; override;
    function  Write(const Data; Len: integer): cardinal;
    procedure WriteStr(const Data: string);
    property OnSessionConnected: TNotifyEvent read FOnSessionConnected write FOnSessionConnected;
    property OnSessionDisconnected: TNotifyEvent read FOnSessionDisconnected write FOnSessionDisconnected;
    property OnDataAvailable: TPipeDataAvailable read FOnDataAvailable write FOnDataAvailable;
    property OnDataSent: TPipeDataSent read FOnDataSent write FOnDataSent;
    property OnError: TPipeError read FOnError write FOnError;
  end;

  TPipeServer = class(TPipe)
  protected
    procedure Execute; override;
  end;

  TPipeClient = class(TPipe)
  private
    FServerName: string;
    procedure SetServerName(const Value: string);
  protected
    procedure Execute; override;
  public
    constructor create(const PipeName: string); override;
    property ServerName: string read FServerName write SetServerName;
  end;

implementation

procedure FileIOCompletionRoutine(ErrorCode, BytesTransferred: dword; Overlap: POverLapped); stdcall;
var
   Pipe: TPipe;
begin
   Pipe := TPipe(Overlap.hEvent);
   Pipe.TriggerDataSent(BytesTransferred);
   Dispose(Overlap);
end;

{ TPipe }

constructor TPipe.create(const PipeName: string);
begin
   inherited Create(True);
   if Pos('\', PipeName) <> 0 then
      raise EPipe.Create('Invalid character in pipe name');
   FPipeName := '\\.\pipe\' + PipeName;
   FPipeHandle := INVALID_HANDLE_VALUE;
   FBufSize := 1024;
   Handle := AllocateHWND(WndProc);
   CloseEvent := CreateEvent(nil, True, False, nil);
end;

destructor TPipe.Destroy;
begin
   if not Suspended then begin
      SetEvent(CloseEvent);
      WaitFor;
   end;
   CloseHandle(CloseEvent);

   if FPipeHandle <> INVALID_HANDLE_VALUE then
      CloseHandle(FPipeHandle);

   DeallocateHWND(Handle);
   inherited;
end;

procedure TPipe.Execute;
var
   WaitHandles: array[0..1] of THandle;
   OverLap: TOverLapped;
   WaitEvent: cardinal;
   Count, BytesLeft: cardinal;
begin
   GetMem(FBuffer, FBufSize + 1);
   FillChar(OverLap, sizeof(OverLapped), 0);
   OverLap.hEvent := CreateEvent(nil, True, False, nil);
   WaitHandles[0] := CloseEvent;
   WaitHandles[1] := OverLap.hEvent;

   try
      while not Terminated do begin
         ReadFile(FPipeHandle, FBuffer^, FBufSize, Count, @Overlap);
         WaitEvent := WaitForMultipleObjects(2, @WaitHandles, False, INFINITE);
         if GetLastError = ERROR_IO_PENDING then
            SetLastError(NO_ERROR);
         case WaitEvent of
            WAIT_OBJECT_0:
               Terminate;
            WAIT_OBJECT_0 + 1: begin
               if not GetOverlappedResult(FPipeHandle, Overlap, Count, False) then
                  Exit;
               if Count = FBufSize then begin
                  if not PeekNamedPipe(FPipeHandle, nil, 0, nil, nil, @BytesLeft) then
                     Exit;
                  if BytesLeft > 0 then begin
                     FBufSize := Count + BytesLeft;
                     ReAllocMem(FBuffer, FBufSize + 1);
                     if not ReadFile(FPipeHandle, FBuffer[Count], BytesLeft, BytesLeft, nil) then
                        Exit;
                     Inc(Count, BytesLeft);
                  end;
               end;
               FBuffer[Count] := #0;
               SendMessage(Handle, WM_DATAAVAILABLE, integer(FBuffer), Count);
            end;
         end;
      end;

   finally
      FreeMem(FBuffer);
      CloseHandle(Overlap.hEvent);
   end;
end;

procedure TPipe.TriggerDataAvailable(Data: PChar; Len: integer);
begin
   if Assigned(FOnDataAvailable) then
      FOnDataAvailable(Self, Data, Len);
end;

procedure TPipe.TriggerDataSent(BytesTransferred: integer);
begin
   if Assigned(FOnDataSent) then
      FOnDataSent(Self, BytesTransferred);
end;

procedure TPipe.TriggerError(Error: integer);
begin
   if Assigned(FOnError) then
      FOnError(Self, Error);
end;

procedure TPipe.TriggerSessionConnected;
begin
   if Assigned(FOnSessionConnected) then
      FOnSessionConnected(Self);
end;

procedure TPipe.TriggerSessionDisconnected;
begin
   if Assigned(FOnSessionDisconnected) then
      FOnSessionDisconnected(Self);
end;

procedure TPipe.WndProc(var MsgRec: TMessage);
begin
   case MsgRec.Msg of
      WM_DATAAVAILABLE:
         TriggerDataAvailable(PChar(MsgRec.WParam), MsgRec.LParam);
      WM_SESSION_CONNECTED:
         TriggerSessionConnected;
      WM_SESSION_DISCONNECTED:
         TriggerSessionDisconnected;
      WM_ERROR:
         TriggerError(MsgRec.WParam);
      else
         MsgRec.Result := DefWindowProc(Handle, MsgRec.Msg, MsgRec.WParam, MsgRec.LParam);
   end;
end;

function TPipe.Write(const Data; Len: integer): cardinal;
var
   OverLap: POverlapped;
begin
   Result := Len;

   New(OverLap);
   FillChar(OverLap^, sizeof(OverLapped), 0);
   OverLap.hEvent := cardinal(Self);

   if not WriteFileEx(FPipeHandle, @Data, Len, Overlap^, @FileIOCompletionRoutine) then begin
      PostMessage(Handle, WM_ERROR, GetLastError, 0);
      Dispose(Overlap);
      Exit;
   end;
   // Use INFINITE instead of 0 for a blocking Write, OnDataSent will fire anyway
   SleepEx(0, True);
end;
// hier

procedure TPipe.WriteStr(const Data: string);
begin
   Write(Data[1], Length(Data));
end;

{ TPipeServer }

procedure TPipeServer.Execute;
var
   WaitHandles: array[0..1] of THandle;
   OverLap: TOverLapped;
   WaitEvent: cardinal;
   SecurityAttr: TSecurityAttributes;
   SecurityDescriptor: PSECURITY_DESCRIPTOR;
begin
   while not Terminated do begin
      if FPipeHandle <> INVALID_HANDLE_VALUE then
         CloseHandle(FPipeHandle);

      FillChar(SecurityAttr, sizeof(TSecurityAttributes), 0);
      GetMem(SecurityDescriptor, SECURITY_DESCRIPTOR_MIN_LENGTH);
      if not InitializeSecurityDescriptor(SecurityDescriptor, SECURITY_DESCRIPTOR_REVISION) then
         TriggerError(GetLastError);
      if not SetSecurityDescriptorDacl(SecurityDescriptor, True, nil, False) then
         TriggerError(GetLastError);
      SecurityAttr.nLength := sizeof(TSecurityAttributes);
      SecurityAttr.bInheritHandle := True;
      SecurityAttr.lpSecurityDescriptor := SecurityDescriptor;

      FPipeHandle:= CreateNamedPipe(PChar(FPipeName), PIPE_ACCESS_DUPLEX or FILE_FLAG_OVERLAPPED,
                                    PIPE_TYPE_MESSAGE or PIPE_WAIT or PIPE_READMODE_MESSAGE,
                                    1, 0, 0, 150, @SecurityAttr);
      FreeMem(SecurityDescriptor);
      
      if FPipeHandle = INVALID_HANDLE_VALUE then begin
         PostMessage(Handle, WM_ERROR, GetLastError, 0); // 231 = ERROR_PIPE_BUSY if pipename already exists
         if WaitForSingleObject(CloseEvent, 1000) = WAIT_OBJECT_0 then
            Terminate;
         Continue;
      end;

      FillChar(OverLap, sizeof(OverLapped), 0);
      OverLap.hEvent := CreateEvent(nil, True, False, nil);
      WaitHandles[0] := CloseEvent;
      WaitHandles[1] := OverLap.hEvent;

      ConnectNamedPipe(FPipeHandle, @Overlap); 
      if GetLastError = ERROR_IO_PENDING then
         SetLastError(NO_ERROR)
      else if GetLastError <> NO_ERROR then
         PostMessage(Handle, WM_ERROR, GetLastError, 0);

      WaitEvent := WaitForMultipleObjects(2, @WaitHandles, False, INFINITE);
      case WaitEvent of
         WAIT_OBJECT_0:
            Terminate;
         WAIT_OBJECT_0 + 1:
            PostMessage(Handle, WM_SESSION_CONNECTED, 0, 0);
      end;
      CloseHandle(Overlap.hEvent);
      inherited;
      if (GetLastError <> 0) and (GetLastError <> ERROR_BROKEN_PIPE) then
         PostMessage(Handle, WM_ERROR, GetLastError, 0);
      PostMessage(Handle, WM_SESSION_DISCONNECTED, 0, 0);
      CloseHandle(FPipeHandle);
      FPipeHandle := INVALID_HANDLE_VALUE;
   end;
end;

{ TPipeClient }

constructor TPipeClient.create(const PipeName: string);
begin
   inherited;
   FServerName := '.';
end;

procedure TPipeClient.Execute;
begin
   while not Terminated do begin
      FPipeHandle:= CreateFile(PChar(FPipeName), GENERIC_READ or GENERIC_WRITE,
                               FILE_SHARE_READ, nil, OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL or FILE_FLAG_OVERLAPPED, 0);
      if FPipeHandle <> INVALID_HANDLE_VALUE then begin
         PostMessage(Handle, WM_SESSION_CONNECTED, 0, 0);
         inherited;
         if (GetLastError <> 0) and (GetLastError <> ERROR_BROKEN_PIPE) then
            PostMessage(Handle, WM_ERROR, GetLastError, 0);
         PostMessage(Handle, WM_SESSION_DISCONNECTED, 0, 0);
         CloseHandle(FPipeHandle);
         FPipeHandle := INVALID_HANDLE_VALUE;
      end else
         PostMessage(Handle, WM_ERROR, GetLastError, 0);
      if WaitForSingleObject(CloseEvent, 1000) = WAIT_OBJECT_0 then
         Terminate;
   end;
end;

procedure TPipeClient.SetServerName(const Value: string);
begin
   if FServerName = Value then
      Exit;
   if not Suspended then
      raise EPipe.Create('Cannot change servername if running');
   FServerName := Value;
   FPipeName := '\\' + FServerName + '\pipe\' + Copy(FPipeName, Pos('pipe\', FPipeName) + 5, Length(FPipeName));
end;

end.
